import uuid
from datetime import datetime, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Farm
from arkindex.process.models import Process, ProcessDataset, ProcessMode, Repository
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Dataset
from arkindex.users.models import Role, User

# Using the fake DB fixtures creation date when needed
FAKE_CREATED = "2020-02-02T01:23:45.678000Z"


class TestProcessDatasets(FixtureAPITestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name="Private corpus")
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = FAKE_CREATED
            cls.private_dataset = cls.private_corpus.datasets.create(
                name="Dead sea scrolls",
                description="Human instrumentality manual",
                creator=cls.user
            )
        cls.test_user = User.objects.create(email="katsuragi@nerv.co.jp", verified_email=True)
        cls.private_corpus.memberships.create(user=cls.test_user, level=Role.Admin.value)

        # Datasets from another corpus
        cls.dataset1, cls.dataset2 = Dataset.objects.filter(corpus=cls.corpus).order_by("name")

        cls.dataset_process = Process.objects.create(
            creator_id=cls.user.id,
            mode=ProcessMode.Dataset,
            corpus_id=cls.private_corpus.id,
            farm=Farm.objects.get(name="Wheat farm")
        )
        cls.process_dataset_1 = ProcessDataset.objects.create(process=cls.dataset_process, dataset=cls.dataset1, sets=cls.dataset1.sets)
        cls.process_dataset_2 = ProcessDataset.objects.create(process=cls.dataset_process, dataset=cls.private_dataset, sets=cls.private_dataset.sets)

        # Control process to check that its datasets are not retrieved
        cls.dataset_process_2 = Process.objects.create(
            creator_id=cls.user.id,
            mode=ProcessMode.Dataset,
            corpus_id=cls.corpus.id
        )
        ProcessDataset.objects.create(process=cls.dataset_process_2, dataset=cls.dataset2, sets=cls.dataset2.sets)

        # For repository process
        cls.repo = Repository.objects.get(url="http://my_repo.fake/workers/worker")
        cls.repo.memberships.create(user=cls.test_user, level=Role.Admin.value)
        cls.rev = cls.repo.revisions.get()

    # List process datasets

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:process-datasets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_process_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:process-datasets", kwargs={"pk": str(uuid.uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_list_process_access_level(self, get_max_level_mock):
        self.private_corpus.memberships.filter(user=self.test_user).delete()
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:process-datasets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_list(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-datasets", kwargs={"pk": self.dataset_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["results"], [
            {
                "id": str(self.process_dataset_2.id),
                "dataset": {
                    "id": str(self.private_dataset.id),
                    "name": "Dead sea scrolls",
                    "description": "Human instrumentality manual",
                    "creator": "Test user",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "corpus_id": str(self.private_corpus.id),
                    "state": "open",
                    "task_id": None,
                    "created": FAKE_CREATED,
                    "updated": FAKE_CREATED
                },
                "sets": ["training", "test", "validation"]
            },
            {
                "id": str(self.process_dataset_1.id),
                "dataset": {
                    "id": str(self.dataset1.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "creator": "Test user",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "corpus_id": str(self.corpus.id),
                    "state": "open",
                    "task_id": None,
                    "created": FAKE_CREATED,
                    "updated": FAKE_CREATED
                },
                "sets": ["training", "test", "validation"]
            }
        ])

    # Create process dataset

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": self.dataset2.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_requires_verified(self):
        unverified_user = User.objects.create(email="email@mail.com")
        self.client.force_login(unverified_user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": self.dataset2.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level")
    def test_create_access_level(self, get_max_level_mock):
        cases = [None, Role.Guest.value, Role.Contributor.value]
        for level in cases:
            with self.subTest(level=level):
                get_max_level_mock.reset_mock()
                get_max_level_mock.return_value = level
                self.client.force_login(self.test_user)

                with self.assertNumQueries(3):
                    response = self.client.post(
                        reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                        data={"sets": self.dataset2.sets}
                    )
                    self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

                self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

                self.assertEqual(get_max_level_mock.call_count, 1)
                self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_create_process_mode(self):
        cases = set(ProcessMode) - {ProcessMode.Dataset, ProcessMode.Local, ProcessMode.Repository}
        for mode in cases:
            with self.subTest(mode=mode):
                self.dataset_process.mode = mode
                self.dataset_process.save()
                self.client.force_login(self.test_user)

                with self.assertNumQueries(5):
                    response = self.client.post(
                        reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                        data={"sets": self.dataset2.sets}
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {"process": ['Datasets can only be added to or removed from processes of mode "dataset".']})

    def test_create_process_mode_local(self):
        self.client.force_login(self.user)
        local_process = Process.objects.get(creator=self.user, mode=ProcessMode.Local)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": local_process.id, "dataset": self.dataset2.id}),
                data={"sets": self.dataset2.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

    def test_create_wrong_process_uuid(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": wrong_id, "dataset": self.dataset2.id}),
                data={"sets": self.dataset2.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"process": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_create_wrong_dataset_uuid(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": wrong_id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"dataset": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    @patch("arkindex.users.managers.BaseACLManager.filter_rights", return_value=Corpus.objects.none())
    def test_create_dataset_access(self, filter_rights_mock):
        new_corpus = Corpus.objects.create(name="NERV")
        new_dataset = new_corpus.datasets.create(name="Eva series", description="We created the Evas from Adam", creator=self.user)
        self.client.force_login(self.test_user)

        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": new_dataset.id}),
                data={"sets": new_dataset.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {"dataset": [f'Invalid pk "{str(new_dataset.id)}" - object does not exist.']})

        self.assertEqual(filter_rights_mock.call_count, 1)
        self.assertEqual(filter_rights_mock.call_args, call(self.test_user, Corpus, Role.Guest.value))

    def test_create_unique(self):
        self.client.force_login(self.test_user)
        self.assertTrue(self.dataset_process.datasets.filter(id=self.dataset1.id).exists())

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": self.dataset1.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"dataset": ["This dataset is already selected in this process."]})

    def test_create_started(self):
        self.client.force_login(self.test_user)
        self.dataset_process.tasks.create(run=0, depth=0, slug="makrout")

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": self.dataset2.sets}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"process": ["Datasets cannot be updated on processes that have already started."]})

    def test_create_default_sets(self):
        self.client.force_login(self.test_user)
        self.assertEqual(ProcessDataset.objects.count(), 3)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse(
                    "api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}
                ),
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(ProcessDataset.objects.count(), 4)
        self.assertTrue(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        self.assertQuerysetEqual(self.dataset_process.datasets.order_by("name"), [
            self.private_dataset,
            self.dataset1,
            self.dataset2
        ])
        created = ProcessDataset.objects.get(process=self.dataset_process.id, dataset=self.dataset2.id)
        self.assertDictEqual(response.json(), {
            "id": str(created.id),
            "dataset": {
                "id": str(self.dataset2.id),
                "name": "Second Dataset",
                "description": "dataset number two",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["training", "test", "validation"]
        })

    def test_create(self):
        self.client.force_login(self.test_user)
        self.assertEqual(ProcessDataset.objects.count(), 3)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": ["validation", "test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(ProcessDataset.objects.count(), 4)
        self.assertTrue(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        self.assertQuerysetEqual(self.dataset_process.datasets.order_by("name"), [
            self.private_dataset,
            self.dataset1,
            self.dataset2
        ])
        created = ProcessDataset.objects.get(process=self.dataset_process.id, dataset=self.dataset2.id)
        self.assertDictEqual(response.json(), {
            "id": str(created.id),
            "dataset": {
                "id": str(self.dataset2.id),
                "name": "Second Dataset",
                "description": "dataset number two",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["validation", "test"]
        })

    def test_create_wrong_sets(self):
        self.client.force_login(self.test_user)
        self.assertEqual(ProcessDataset.objects.count(), 3)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": ["Unit-01"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process.id, dataset=self.dataset2.id).exists())
        self.assertDictEqual(response.json(), {"sets": ["The specified sets must all exist in the specified dataset."]})

    # Update process dataset

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_requires_verified(self):
        unverified_user = User.objects.create(email="email@mail.com")
        self.client.force_login(unverified_user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_access_level(self):
        cases = [None, Role.Guest, Role.Contributor]
        for level in cases:
            with self.subTest(level=level):
                self.private_corpus.memberships.filter(user=self.test_user).delete()
                if level:
                    self.private_corpus.memberships.create(user=self.test_user, level=level.value)
                self.client.force_login(self.test_user)
                with self.assertNumQueries(4):
                    response = self.client.put(
                        reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                        data={"sets": ["test"]}
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_process_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_dataset_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_m2m_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.process_dataset_1.id),
            "dataset": {
                "id": str(self.dataset1.id),
                "name": "First Dataset",
                "description": "dataset number one",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["test"]
        })

    def test_update_wrong_sets(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["Unit-01", "Unit-02"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["The specified sets must all exist in the specified dataset."]})

    def test_update_unique_sets(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test", "test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Sets must be unique."]})

    def test_update_started_process(self):
        self.dataset_process.tasks.create(
            run=0,
            depth=0,
            slug="task",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
        )
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"process": ["Datasets cannot be updated on processes that have already started."]})

    def test_update_only_sets(self):
        """
        Non "sets" fields in the update request are ignored
        """
        self.client.force_login(self.test_user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"process": str(self.dataset_process_2.id), "dataset": str(self.dataset2.id), "sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.process_dataset_1.id),
            "dataset": {
                "id": str(self.dataset1.id),
                "name": "First Dataset",
                "description": "dataset number one",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["test"]
        })

    # Partial update process dataset

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_requires_verified(self):
        unverified_user = User.objects.create(email="email@mail.com")
        self.client.force_login(unverified_user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_access_level(self):
        cases = [None, Role.Guest, Role.Contributor]
        for level in cases:
            with self.subTest(level=level):
                self.private_corpus.memberships.filter(user=self.test_user).delete()
                if level:
                    self.private_corpus.memberships.create(user=self.test_user, level=level.value)
                self.client.force_login(self.test_user)
                with self.assertNumQueries(4):
                    response = self.client.patch(
                        reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                        data={"sets": ["test"]}
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_partial_update_process_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update_dataset_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update_m2m_does_not_exist(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_partial_update(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.process_dataset_1.id),
            "dataset": {
                "id": str(self.dataset1.id),
                "name": "First Dataset",
                "description": "dataset number one",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["test"]
        })

    def test_partial_update_wrong_sets(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["Unit-01", "Unit-02"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["The specified sets must all exist in the specified dataset."]})

    def test_partial_update_unique_sets(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test", "test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Sets must be unique."]})

    def test_partial_update_started_process(self):
        self.dataset_process.tasks.create(
            run=0,
            depth=0,
            slug="task",
            expiry=datetime(1970, 1, 1, tzinfo=timezone.utc),
        )
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"process": ["Datasets cannot be updated on processes that have already started."]})

    def test_partial_update_only_sets(self):
        """
        Non "sets" fields in the partial update request are ignored
        """
        self.client.force_login(self.test_user)
        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
                data={"process": str(self.dataset_process_2.id), "dataset": str(self.dataset2.id), "sets": ["test"]}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.process_dataset_1.id),
            "dataset": {
                "id": str(self.dataset1.id),
                "name": "First Dataset",
                "description": "dataset number one",
                "creator": "Test user",
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "corpus_id": str(self.corpus.id),
                "state": "open",
                "task_id": None,
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED
            },
            "sets": ["test"]
        })

    # Destroy process dataset

    def test_destroy_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.private_dataset.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_process_does_not_exist(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": wrong_id, "dataset": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"process": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_destroy_dataset_does_not_exist(self):
        self.client.force_login(self.test_user)
        wrong_id = uuid.uuid4()
        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": wrong_id})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"dataset": [f'Invalid pk "{str(wrong_id)}" - object does not exist.']})

    def test_destroy_not_found(self):
        self.assertFalse(self.dataset_process.datasets.filter(id=self.dataset2.id).exists())
        self.client.force_login(self.test_user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level", return_value=None)
    def test_destroy_process_access_level(self, get_max_level_mock):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to this process."})

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.test_user, self.private_corpus))

    def test_destroy_no_dataset_access_requirement(self):
        new_corpus = Corpus.objects.create(name="NERV")
        new_dataset = new_corpus.datasets.create(name="Eva series", description="We created the Evas from Adam", creator=self.user)
        ProcessDataset.objects.create(process=self.dataset_process, dataset=new_dataset, sets=new_dataset.sets)
        self.assertTrue(ProcessDataset.objects.filter(process=self.dataset_process, dataset=new_dataset).exists())
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": new_dataset.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process, dataset=new_dataset).exists())

    def test_destroy_process_mode(self):
        cases = set(ProcessMode) - {ProcessMode.Dataset, ProcessMode.Local, ProcessMode.Repository}
        for mode in cases:
            with self.subTest(mode=mode):
                self.dataset_process.mode = mode
                self.dataset_process.save()
                self.client.force_login(self.test_user)

                with self.assertNumQueries(4):
                    response = self.client.delete(
                        reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset2.id}),
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertEqual(response.json(), {"process": ['Datasets can only be added to or removed from processes of mode "dataset".']})

    def test_destroy_process_mode_local(self):
        self.client.force_login(self.user)
        local_process = Process.objects.get(creator=self.user, mode=ProcessMode.Local)
        with self.assertNumQueries(3):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": local_process.id, "dataset": self.dataset2.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have admin access to this process."})

    def test_destroy_started(self):
        self.client.force_login(self.test_user)
        self.dataset_process.tasks.create(run=0, depth=0, slug="makrout")

        with self.assertNumQueries(4):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"process": ["Datasets cannot be updated on processes that have already started."]})

    def test_destroy(self):
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process, dataset=self.dataset1).exists())

    def test_destroy_sets_agnostic(self):
        """
        When deleting a process dataset, it doesn't matter what its sets are as there cannot be two process datasets
        with the same process and dataset, whatever the sets are.
        """
        self.process_dataset_1.sets = ["test"]
        self.process_dataset_1.save()
        self.client.force_login(self.test_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:process-dataset", kwargs={"process": self.dataset_process.id, "dataset": self.dataset1.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(ProcessDataset.objects.filter(process=self.dataset_process, dataset=self.dataset1).exists())
